"""Frozen Dict"""

from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

__all__ = ("FrozenDict",)

V = TypeVar("V", bound=Any)
SelfT = TypeVar("SelfT", bound="FrozenDict[Any]")


class FrozenDict(Mapping[str, V]):
    """Frozen Mapping.

    An immutable mapping of string -> Any type
    """

    __slots__ = ("__hash", "__data")

    def __getitem__(self, k: str) -> V:
        return self.__data[k]

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__data)

    def __init__(
        self,
        seq: Union[Mapping[str, V], Iterable[Tuple[str, V]], None] = None,
        **kwargs: V,
    ) -> None:
        d = dict(seq, **kwargs) if seq is not None else dict(**kwargs)

        self.__data: Dict[str, V] = d
        self.__hash: Optional[int] = None

    def copy(self: SelfT, **kwargs: V) -> SelfT:
        """Copy dict, replacing values according to kwargs.

        >>> fd = FrozenDict(a=1)
        >>> fd["a"]
        1
        >>> fd1 = fd.copy(a=2)
        >>> fd1["a"]
        2
        >>> fd["a"]
        1
        """
        return self.__class__(self.__data, **kwargs)

    def __hash__(self) -> int:
        if self.__hash is None:
            self.__hash = hash(tuple(self.items()))

        return self.__hash



"""
IRC string utils
"""

import operator
import string
from typing import (
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Protocol,
    SupportsIndex,
    Tuple,
    Union,
)

__all__ = ("Casemap", "RFC1459", "STRICT_RFC1459", "ASCII", "String")


class Casemap(NamedTuple):
    """String case-map

    Represents a mapping of lower to upper case letters
    """

    lower: str
    upper: str

    @property
    def lower_table(self) -> Dict[int, int]:
        """The lower->upper translation table"""
        return str.maketrans(self.lower, self.upper)

    @property
    def upper_table(self) -> Dict[int, int]:
        """The upper->lower table"""
        return str.maketrans(self.upper, self.lower)


RFC1459 = Casemap(
    "".join(map(chr, range(65, 95))), "".join(map(chr, range(97, 127)))
)
STRICT_RFC1459 = Casemap(
    "".join(map(chr, range(65, 94))), "".join(map(chr, range(97, 126)))
)
ASCII = Casemap(
    "".join(map(chr, range(65, 91))), "".join(map(chr, range(97, 123)))
)


class TranslateTable(Protocol):
    def __getitem__(self, item: int, /) -> Union[str, int, None]:
        raise NotImplementedError


class String(str):
    """Case-insensitive string"""

    __slots__ = ("_casemap",)

    _casemap: Casemap

    def _wrap(self, value: str) -> "String":
        return self.__class__.__new__(self.__class__, value, self.casemap)

    def __new__(
        cls, value: str = "", casemap: Optional[Casemap] = None
    ) -> "String":
        o = str.__new__(cls, value)
        o._casemap = casemap or ASCII  # noqa: SLF001
        return o

    def __hash__(self) -> int:
        return hash(str(self.lower()))

    def __internal_cmp(
        self, other: object, cmp: Callable[[str, str], bool]
    ) -> bool:
        if isinstance(other, String):
            return cmp(str(self.casefold()), str(other.casefold()))

        if isinstance(other, str):
            return cmp(self, self._wrap(other))

        return NotImplemented

    def __lt__(self, other: str) -> bool:
        return self.__internal_cmp(other, operator.lt)

    def __le__(self, other: str) -> bool:
        return self.__internal_cmp(other, operator.le)

    def __eq__(self, other: object) -> bool:
        return self.__internal_cmp(other, operator.eq)

    def __ne__(self, other: object) -> bool:
        return self.__internal_cmp(other, operator.ne)

    def __gt__(self, other: str) -> bool:
        return self.__internal_cmp(other, operator.gt)

    def __ge__(self, other: str) -> bool:
        return self.__internal_cmp(other, operator.ge)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, str):
            return False

        return self._wrap(item).casefold() in str(self.casefold())

    def __getitem__(self, item: Union[SupportsIndex, slice]) -> "String":
        return self._wrap(super().__getitem__(item))

    def translate(self, table: TranslateTable) -> "String":
        return self._wrap(super().translate(table))

    @property
    def _capitalize(self) -> "String":
        return self[:1].upper() + self[1:].lower()

    def __add__(self, other: str) -> "String":
        return self._wrap(str(self) + other)

    def capitalize(self) -> "String":
        return self._capitalize

    @property
    def _casefold(self) -> "String":
        return self.lower()

    def casefold(self) -> "String":
        return self._casefold

    @property
    def _lower(self) -> "String":
        return self.translate(self.casemap.lower_table)

    def lower(self) -> "String":
        return self._lower

    @property
    def _upper(self) -> "String":
        return self.translate(self.casemap.upper_table)

    def upper(self) -> "String":
        return self._upper

    def count(
        self,
        sub: str,
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> int:
        return str(self.casefold()).count(
            self._wrap(sub).casefold(), start, end
        )

    def startswith(
        self,
        prefix: Union[str, Tuple[str, ...]],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        prefix_list: Tuple[str, ...]
        prefix_list = (prefix,) if isinstance(prefix, str) else prefix

        mapped_list = tuple(self._wrap(p).casefold() for p in prefix_list)

        return str(self.casefold()).startswith(mapped_list, start, end)

    def endswith(
        self,
        suffix: Union[str, Tuple[str, ...]],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        suffix_list: Tuple[str, ...]
        suffix_list = (suffix,) if isinstance(suffix, str) else suffix

        mapped_list = tuple(self._wrap(p).casefold() for p in suffix_list)

        return str(self.casefold()).endswith(mapped_list, start, end)

    def find(
        self,
        sub: str,
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> int:
        return str(self.casefold()).find(self._wrap(sub).casefold(), start, end)

    def rfind(
        self,
        sub: str,
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> int:
        return str(self.casefold()).rfind(
            self._wrap(sub).casefold(), start, end
        )

    def index(
        self,
        sub: str,
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> int:
        return str(self.casefold()).index(
            self._wrap(sub).casefold(), start, end
        )

    def rindex(
        self,
        sub: str,
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> int:
        return str(self.casefold()).rindex(
            self._wrap(sub).casefold(), start, end
        )

    def partition(self, sep: str) -> Tuple["String", "String", "String"]:
        pos = self.find(sep)
        if pos < 0:
            return self, self._wrap(""), self._wrap("")

        return self[:pos], self[pos : pos + len(sep)], self[pos + len(sep) :]

    def rpartition(self, sep: str) -> Tuple["String", "String", "String"]:
        pos = self.rfind(sep)
        if pos < 0:
            return self._wrap(""), self._wrap(""), self

        return self[:pos], self[pos : pos + len(sep)], self[pos + len(sep) :]

    def replace(
        self, old: str, new: str, count: SupportsIndex = -1
    ) -> "String":
        raise NotImplementedError

    def strip(self, chars: Optional[str] = None) -> "String":
        return self.lstrip(chars).rstrip(chars)

    def lstrip(self, chars: Optional[str] = None) -> "String":
        if chars is None:
            chars = string.whitespace

        chars = self._wrap(chars)

        start = 0
        while start < len(self) and self[start] in chars:
            start += 1

        return self[start:]

    def rstrip(self, chars: Optional[str] = None) -> "String":
        if chars is None:
            chars = string.whitespace

        chars = self._wrap(chars)

        end = len(self)
        while end > 0 and self[end - 1] in chars:
            end -= 1

        return self[:end]

    @property
    def _swapcase(self) -> "String":
        trans = self.casemap.lower_table.copy()
        trans.update(self.casemap.upper_table)

        return self.translate(trans)

    def swapcase(self) -> "String":
        return self._swapcase

    def title(self) -> str:
        raise NotImplementedError

    def split(
        self, sep: Optional[str] = None, maxsplit: SupportsIndex = -1
    ) -> List[str]:
        raise NotImplementedError

    def rsplit(
        self, sep: Optional[str] = None, maxsplit: SupportsIndex = -1
    ) -> List[str]:
        raise NotImplementedError

    @property
    def casemap(self) -> Casemap:
        """Casemap associated with this string"""
        return self._casemap



"""
Test constructing message objects
"""

from irclib.parser import Message


def test_param_construct() -> None:
    """Test constructing Message objects"""
    msg = Message(None, None, "PRIVMSG", "#channel", "Message thing")
    assert str(msg) == "PRIVMSG #channel :Message thing"

    msg = Message(None, None, "PRIVMSG", ["#channel", "Message thing"])
    assert str(msg) == "PRIVMSG #channel :Message thing"

    msg = Message(None, None, "PRIVMSG", ["#channel", ":Message thing"])
    assert str(msg) == "PRIVMSG #channel ::Message thing"

    msg = Message(None, None, "PRIVMSG", [""])
    assert str(msg) == "PRIVMSG :"



"""Test IRC parser"""

from typing import Dict, List, Optional, Tuple, Type, TypedDict

import parser_tests.data
import pytest

from irclib.parser import (
    Cap,
    CapList,
    Message,
    MessageTag,
    ParamList,
    Parseable,
    Prefix,
    TagList,
)
from irclib.util.string import ASCII, String


class MsgAtoms(TypedDict):
    tags: Dict[str, str]
    source: str
    verb: str
    params: List[str]


class MsgSplitCase(TypedDict):
    input: str
    atoms: MsgAtoms


class UserHostAtoms(TypedDict):
    nick: str
    user: str
    host: str


class UserHostSplitCase(TypedDict):
    source: str
    atoms: UserHostAtoms


class MsgJoinCase(TypedDict):
    atoms: MsgAtoms
    matches: List[str]


def test_line() -> None:
    """Test parsing a single IRC message"""
    assert Message.parse("COMMAND") == "COMMAND"
    assert Message.parse("command") == "COMMAND"

    msg1 = Message.parse("PRIVMSG")

    assert msg1.command == "PRIVMSG"

    msg2 = Message.parse(
        r"@test=data;test1=more\sdata :nick!user@host COMMAND arg1 arg2 :trailing text"
    )

    assert msg2.prefix
    assert msg2.prefix.host == "host"


class TestCap:
    """Test Cap class"""

    @pytest.mark.parametrize(
        ("name", "value", "expected"),
        [
            ("capname", "somevalue", "capname=somevalue"),
            ("capname", None, "capname"),
        ],
    )
    def test_str(self, name: str, value: Optional[str], expected: str) -> None:
        """Test string conversion"""
        c = Cap(name, value)
        assert str(c) == expected

    @pytest.mark.parametrize(("name", "value"), [("a", "b"), ("foo", "bar")])
    def test_eq(self, name: str, value: str) -> None:
        """Test equals"""
        assert Cap(name, value) == Cap(name, value)

    @pytest.mark.parametrize(
        ("name", "value", "string"),
        [("a", "b", "a=b"), ("foo", "bar", "foo=bar")],
    )
    def test_eq_str(self, name: str, value: str, string: str) -> None:
        """Test equals string"""
        assert Cap(name, value) == string
        assert string == Cap(name, value)

    @pytest.mark.parametrize(
        ("name", "value", "other_name", "other_value"),
        [
            ("a", "b", "c", "b"),
            ("a", "b", "a", "c"),
            ("a", "b", "c", "d"),
            ("foo", "bar", "baz", "zing"),
            ("foo", "bar", "baz", None),
            ("foo", None, "baz", None),
            ("foo", None, "baz", "zing"),
            ("foo", "bar", "foo", "zing"),
            ("foo", "bar", "foo", None),
            ("foo", None, "foo", "zing"),
        ],
    )
    def test_ne(
        self,
        name: str,
        value: Optional[str],
        other_name: str,
        other_value: Optional[str],
    ) -> None:
        """Test not-equals"""
        assert Cap(name, value) != Cap(other_name, other_value)
        assert Cap(other_name, other_value) != Cap(name, value)

    @pytest.mark.parametrize(
        ("name", "value", "string"),
        [
            ("a", "b", "c=b"),
            ("a", "b", "a=c"),
            ("a", "b", "c=d"),
            ("foo", "bar", "baz=zing"),
            ("foo", "bar", "foo=zing"),
            ("foo", "bar", "foo"),
            ("foo", "bar", "foo=baz"),
            ("foo", None, "foo=baz"),
            ("foo", None, "baz"),
        ],
    )
    def test_ne_str(self, name: str, value: Optional[str], string: str) -> None:
        """Test not-equals string"""
        assert Cap(name, value) != string
        assert string != Cap(name, value)

    @pytest.mark.parametrize(("obj", "other"), [(Cap("foo"), 1)])
    def test_no_cmp(self, obj: Cap, other: object) -> None:
        """Test not-equals"""
        assert obj != other
        assert other != obj

        assert not obj == other
        assert not other == obj

    @pytest.mark.parametrize(
        ("text", "name", "value"),
        [("vendor.example.org/cap-name", "vendor.example.org/cap-name", None)],
    )
    def test_parse(self, text: str, name: str, value: Optional[str]) -> None:
        """Test parsing string"""
        cap = Cap.parse(text)
        assert cap.name == name
        assert cap.value == value


class TestCapList:
    """Test parsing a list of CAPs"""

    @pytest.mark.parametrize(
        ("text", "expected"),
        [
            (
                "blah blah-blah cap-1 test-cap=value-data",
                (
                    ("blah", None),
                    ("blah-blah", None),
                    ("cap-1", None),
                    ("test-cap", "value-data"),
                ),
            ),
            (
                "blah blah-blah cap-1 test-cap=value-data ",
                (
                    ("blah", None),
                    ("blah-blah", None),
                    ("cap-1", None),
                    ("test-cap", "value-data"),
                ),
            ),
            (
                ":blah blah-blah cap-1 test-cap=value-data",
                (
                    ("blah", None),
                    ("blah-blah", None),
                    ("cap-1", None),
                    ("test-cap", "value-data"),
                ),
            ),
            (
                ":blah blah-blah cap-1 test-cap=value-data ",
                (
                    ("blah", None),
                    ("blah-blah", None),
                    ("cap-1", None),
                    ("test-cap", "value-data"),
                ),
            ),
            ("", ()),
        ],
    )
    def test_parse(
        self, text: str, expected: Tuple[Tuple[str, Optional[str]], ...]
    ) -> None:
        """Test string parsing"""
        parsed = CapList.parse(text)
        assert len(parsed) == len(expected)
        for (name, value), actual in zip(expected, parsed):
            assert actual.name == name
            assert actual.value == value

    @pytest.mark.parametrize("caps", [[], [Cap("a"), Cap("b", "c")]])
    def test_eq_list(self, caps: List[Cap]) -> None:
        """Test equals list"""
        assert CapList(caps) == caps
        assert caps == CapList(caps)

    @pytest.mark.parametrize("caps", [[], [Cap("a"), Cap("b", "c")]])
    def test_ne_list(self, caps: List[Cap]) -> None:
        """Test not-equals list"""
        b = CapList(caps) != caps
        assert not b
        b1 = caps != CapList(caps)
        assert not b1

    @pytest.mark.parametrize(
        ("caps", "text"),
        [
            ([], ""),
            ([Cap("a"), Cap("b", "c")], "a b=c"),
            ([Cap("a"), Cap("b", "c")], "a b=c "),
        ],
    )
    def test_eq_str(self, caps: List[Cap], text: str) -> None:
        """Test equals strings"""
        assert CapList(caps) == text
        assert text == CapList(caps)

    @pytest.mark.parametrize(
        ("caps", "text"),
        [
            ([], ""),
            ([Cap("a"), Cap("b", "c")], "a b=c"),
            ([Cap("a"), Cap("b", "c")], "a b=c "),
        ],
    )
    def test_ne_str(self, caps: List[Cap], text: str) -> None:
        """Test not-equals strings"""
        b = CapList(caps) != text
        assert not b
        b1 = text != CapList(caps)
        assert not b1

    @pytest.mark.parametrize(
        ("obj", "other"), [(CapList(), None), (CapList(), 0)]
    )
    def test_no_cmp(self, obj: CapList, other: object) -> None:
        """Test not-equals"""
        assert obj != other
        assert other != obj

        assert not obj == other
        assert not other == obj

    @pytest.mark.parametrize(
        ("obj", "text"),
        [
            (CapList([Cap("foo")]), "foo"),
            (CapList([Cap("foo"), Cap("bar")]), "foo bar"),
            (CapList([Cap("foo"), Cap("bar=baz")]), "foo bar=baz"),
        ],
    )
    def test_str(self, obj: CapList, text: str) -> None:
        """Test string conversion"""
        assert str(obj) == text
        assert text == str(obj)


class TestMessageTag:
    """Test parsing a single message tag"""

    @pytest.mark.parametrize(
        ("text", "name", "value"),
        [
            ("a=b", "a", "b"),
            ("test/blah=", "test/blah", ""),
            (r"blah=aa\r\n\:\\", "blah", "aa\r\n;\\"),
            (r"blah=aa\r\n\s\:\\", "blah", "aa\r\n ;\\"),
            (r"blah=\s", "blah", " "),
            ("blah=\\", "blah", ""),
        ],
    )
    def test_parse(self, text: str, name: str, value: Optional[str]) -> None:
        """Test parsing a string"""
        tag = MessageTag.parse(text)
        assert tag.name == name
        assert tag.value == value

    @pytest.mark.parametrize(("name", "value"), [("a", None), ("a", "b")])
    def test_eq(self, name: str, value: Optional[str]) -> None:
        """Test equals"""
        assert MessageTag(name, value) == MessageTag(name, value)

    @pytest.mark.parametrize(("name", "value"), [("a", None), ("a", "b")])
    def test_ne(self, name: str, value: Optional[str]) -> None:
        """Test not-equals"""
        b = MessageTag(name, value) != MessageTag(name, value)
        assert not b

    @pytest.mark.parametrize(
        ("name", "value", "text"),
        [("foo", None, "foo"), ("foo", "bar", "foo=bar")],
    )
    def test_eq_str(self, name: str, value: Optional[str], text: str) -> None:
        """Test equals string"""
        assert MessageTag(name, value) == text
        assert text == MessageTag(name, value)

    @pytest.mark.parametrize(
        ("name", "value", "text"),
        [("foo", None, "foo"), ("foo", "bar", "foo=bar")],
    )
    def test_ne_str(self, name: str, value: Optional[str], text: str) -> None:
        """Test not-equals string"""
        b = MessageTag(name, value) != text
        assert not b
        b1 = text != MessageTag(name, value)
        assert not b1

    @pytest.mark.parametrize(
        ("obj", "other"), [(MessageTag(""), None), (MessageTag(""), 1)]
    )
    def test_no_cmp(self, obj: MessageTag, other: object) -> None:
        """Test not-equals other types"""
        assert obj != other
        assert other != obj

        assert not obj == other
        assert not other == obj


class TestTagList:
    """Test parsing a tag list"""

    @pytest.mark.parametrize(
        ("text", "tags"),
        [
            ("a;", [("a", "")]),
            ("a=1;", [("a", "1")]),
            ("a=1;b;c", [("a", "1"), ("b", ""), ("c", "")]),
            ("a=1;b;c;", [("a", "1"), ("b", ""), ("c", "")]),
            ("", []),
            (";", []),
            (r"a=ab\r\s\n\:\\;b=abc", [("a", "ab\r \n;\\"), ("b", "abc")]),
        ],
    )
    def test_parse(self, text: str, tags: List[Tuple[str, str]]) -> None:
        """Test parsing from string"""
        tag_list = TagList.parse(text)

        assert len(tag_list) == len(tags)

        for name, value in tags:
            tag = tag_list[name]
            assert tag.name == name
            assert tag.value == value

    @pytest.mark.parametrize(
        "tags", [[], [MessageTag("a")], [MessageTag("b", "c")]]
    )
    def test_eq(self, tags: List[MessageTag]) -> None:
        """Test equals"""
        assert TagList(tags) == TagList(tags)

    @pytest.mark.parametrize(
        "tags", [[], [MessageTag("a")], [MessageTag("b", "c")]]
    )
    def test_ne(self, tags: List[MessageTag]) -> None:
        """Test not-equals"""
        b = TagList(tags) != TagList(tags)
        assert not b

    @pytest.mark.parametrize(
        "tags", [[], [MessageTag("a")], [MessageTag("b", "c")]]
    )
    def test_eq_list(self, tags: List[MessageTag]) -> None:
        """Test equals list"""
        assert TagList(tags) == tags
        assert tags == TagList(tags)

    @pytest.mark.parametrize(
        "tags", [[], [MessageTag("a")], [MessageTag("b", "c")]]
    )
    def test_ne_list(self, tags: List[MessageTag]) -> None:
        """Test not-equals list"""
        b = TagList(tags) != tags
        assert not b
        b1 = tags != TagList(tags)
        assert not b1

    @pytest.mark.parametrize(
        ("obj", "other"),
        [
            (TagList(), {}),
            (TagList([MessageTag("a")]), {"a": None}),
            (TagList([MessageTag("b", "c")]), {"b": "c"}),
        ],
    )
    def test_eq_dict(
        self, obj: TagList, other: Dict[str, Optional[str]]
    ) -> None:
        """Test equals dict"""
        assert obj == other
        assert other == obj

    @pytest.mark.parametrize(
        ("obj", "other"),
        [
            (TagList(), {}),
            (TagList([MessageTag("a")]), {"a": None}),
            (TagList([MessageTag("b", "c")]), {"b": "c"}),
        ],
    )
    def test_ne_dict(
        self, obj: TagList, other: Dict[str, Optional[str]]
    ) -> None:
        """Test not-equals dict"""
        b = obj != other
        assert not b
        b1 = other != obj
        assert not b1

    @pytest.mark.parametrize(
        ("tags", "text"),
        [([], ""), ([MessageTag("a")], "a"), ([MessageTag("b", "c")], "b=c")],
    )
    def test_eq_str(self, tags: List[MessageTag], text: str) -> None:
        """Test equals strings"""
        assert TagList(tags) == text
        assert text == TagList(tags)

    @pytest.mark.parametrize(
        ("tags", "text"),
        [([], ""), ([MessageTag("a")], "a"), ([MessageTag("b", "c")], "b=c")],
    )
    def test_ne_str(self, tags: List[MessageTag], text: str) -> None:
        """Test not-equals strings"""
        b = TagList(tags) != text
        assert not b
        b1 = text != TagList(tags)
        assert not b1

    @pytest.mark.parametrize(
        ("obj", "other"), [(TagList(), 0), (TagList(), None)]
    )
    def test_no_cmp(self, obj: TagList, other: object) -> None:
        """Test not-equals other types"""
        assert obj != other
        assert other != obj

        assert not obj == other
        assert not other == obj


class TestPrefix:
    """Test parsing a prefix"""

    @pytest.mark.parametrize(
        ("text", "nick", "user", "host"),
        [
            ("", "", "", ""),
            (":", "", "", ""),
            ("somenick", "somenick", "", ""),
            ("somenick!user", "somenick", "user", ""),
            ("somenick@host", "somenick", "", "host"),
            ("somenick!user@host", "somenick", "user", "host"),
            (":somenick", "somenick", "", ""),
            (":somenick!user", "somenick", "user", ""),
            (":somenick@host", "somenick", "", "host"),
            (":somenick!user@host", "somenick", "user", "host"),
            (":nick!@", "nick", "", ""),
            (":nick!user@", "nick", "user", ""),
            (":nick!@host", "nick", "", "host"),
            (":!@", "", "", ""),
            (":!user@", "", "user", ""),
            (":!@host", "", "", "host"),
            (":!user@host", "", "user", "host"),
        ],
    )
    def test_parse(self, text: str, nick: str, user: str, host: str) -> None:
        """Test parsing a string"""
        p = Prefix.parse(text)

        assert p.nick == nick
        assert p.user == user
        assert p.ident == user
        assert p.host == host

    @pytest.mark.parametrize(
        ("nick", "user", "host"),
        [
            ("nick", None, None),
            ("nick", "user", "host"),
            ("nick", None, "host"),
            ("nick", "user", None),
        ],
    )
    def test_eq(
        self, nick: str, user: Optional[str], host: Optional[str]
    ) -> None:
        """Test equals"""
        assert Prefix(nick, user, host) == Prefix(nick, user, host)

    @pytest.mark.parametrize(
        ("nick", "user", "host"),
        [
            ("nick", None, None),
            ("nick", "user", "host"),
            ("nick", None, "host"),
            ("nick", "user", None),
        ],
    )
    def test_ne(
        self, nick: str, user: Optional[str], host: Optional[str]
    ) -> None:
        """Test not-equals"""
        b = Prefix(nick, user, host) != Prefix(nick, user, host)
        assert not b

    @pytest.mark.parametrize(
        ("text", "nick", "user", "host"),
        [
            ("nick", "nick", None, None),
            ("nick!user@host", "nick", "user", "host"),
            ("nick@host", "nick", None, "host"),
            ("nick!user", "nick", "user", None),
        ],
    )
    def test_eq_str(
        self,
        text: str,
        nick: Optional[str],
        user: Optional[str],
        host: Optional[str],
    ) -> None:
        """Test equals string"""
        assert Prefix(nick, user, host) == text
        assert text == Prefix(nick, user, host)

    @pytest.mark.parametrize(
        ("text", "nick", "user", "host"),
        [
            ("nick", "nick", None, None),
            ("nick!user@host", "nick", "user", "host"),
            ("nick@host", "nick", None, "host"),
            ("nick!user", "nick", "user", None),
        ],
    )
    def test_ne_str(
        self,
        text: str,
        nick: Optional[str],
        user: Optional[str],
        host: Optional[str],
    ) -> None:
        """Test not-equals string comparisons"""
        b = Prefix(nick, user, host) != text
        assert not b
        b1 = text != Prefix(nick, user, host)
        assert not b1

    @pytest.mark.parametrize(
        ("obj", "other"),
        [(Prefix(""), 0), (Prefix(""), None), (Prefix(""), ())],
    )
    def test_no_cmp(self, obj: Prefix, other: object) -> None:
        """Test not-equals"""
        assert obj != other
        assert other != obj

        assert not obj == other
        assert not other == obj

    @pytest.mark.parametrize(
        ("obj", "nick", "user", "host"),
        [
            (Prefix("nick"), "nick", "", ""),
            (Prefix("nick", "user", "host"), "nick", "user", "host"),
            (Prefix("nick", host="host"), "nick", "", "host"),
            (Prefix("nick", "user"), "nick", "user", ""),
        ],
    )
    def test_unpack(
        self,
        obj: Prefix,
        nick: Optional[str],
        user: Optional[str],
        host: Optional[str],
    ) -> None:
        """Test unpacking Prefix"""
        n, u, h = obj
        assert (n, u, h) == (nick, user, host)

    @pytest.mark.parametrize(
        ("nick", "user", "host"),
        [
            ("nick", "", ""),
            ("nick", None, None),
            ("nick", "user", "host"),
            ("nick", "", "host"),
            ("nick", "user", ""),
            ("nick", None, "host"),
            ("nick", "user", None),
            ("", "user", ""),
            ("", "user", "host"),
            ("", "", "host"),
            (None, "user", None),
            (None, "user", "host"),
            (None, None, "host"),
        ],
    )
    def test_bool(
        self, nick: Optional[str], user: Optional[str], host: Optional[str]
    ) -> None:
        """Test cases where bool(Prefix) == True"""
        assert Prefix(nick, user, host)

    @pytest.mark.parametrize(
        ("nick", "user", "host"),
        [
            ("", "", ""),
            (None, None, None),
            ("", None, None),
            (None, "", None),
            (None, None, ""),
            ("", "", None),
            ("", None, ""),
            (None, "", ""),
        ],
    )
    def test_bool_false(
        self, nick: Optional[str], user: Optional[str], host: Optional[str]
    ) -> None:
        """Test cases where bool(Prefix) == False"""
        assert not Prefix(nick, user, host)


class TestParamList:
    """Test parsing a parameter list"""

    @pytest.mark.parametrize(
        ("obj", "text"),
        [
            (ParamList(), ""),
            (ParamList("a"), "a"),
            (ParamList("a", "b"), "a b"),
            (ParamList("a", "b", has_trail=True), "a :b"),
            (ParamList("a", "b", ":c"), "a b ::c"),
            (ParamList("a", "b", "c "), "a b :c "),
            (ParamList("a", "b", ""), "a b :"),
        ],
    )
    def test_str(self, obj: ParamList, text: str) -> None:
        """Test string conversion"""
        assert str(obj) == text

    @pytest.mark.parametrize(
        "params",
        [[], [""], ["a"], ["a", "b"], ["a", ":b"], ["a", "b "], ["a", ""]],
    )
    def test_eq(self, params: List[str]) -> None:
        """Test equals"""
        assert ParamList(*params) == ParamList(*params)
        assert ParamList(*params) == ParamList.from_list(params)
        assert ParamList.from_list(params) == ParamList(*params)

    @pytest.mark.parametrize(
        "params",
        [[], [""], ["a"], ["a", "b"], ["a", ":b"], ["a", "b "], ["a", ""]],
    )
    def test_ne(self, params: List[str]) -> None:
        """Test not-equals"""
        b = ParamList(*params) != ParamList(*params)
        assert not b

    @pytest.mark.parametrize(
        ("obj", "other"), [(ParamList(), 0), (ParamList(), None)]
    )
    def test_no_cmp(self, obj: ParamList, other: object) -> None:
        """Test not-equals"""
        assert obj != other
        assert other != obj

        assert not obj == other
        assert not other == obj


class TestMessage:
    """Test parsing an entire IRC message"""

    def test_parse_bytes(self) -> None:
        """Test parsing bytes"""
        line = Message.parse(b"COMMAND some params :and stuff")
        assert line.command == "COMMAND"
        assert line.parameters == ["some", "params", "and stuff"]

    @pytest.mark.parametrize(
        ("obj", "text"),
        [
            (Message(None, None, None), ""),
            (Message(None, None, None, None), ""),
            (Message(None, None, None, []), ""),
            (Message(None, None, "COMMAND"), "COMMAND"),
            (Message(["a=b"], None, "COMMAND"), "@a=b COMMAND"),
            (Message([MessageTag("a", "b")], None, "COMMAND"), "@a=b COMMAND"),
            (Message({"a": "b"}, None, "COMMAND"), "@a=b COMMAND"),
            (Message({"a": "b"}, "nick", "COMMAND"), "@a=b :nick COMMAND"),
            (Message(None, ("nick",), "COMMAND"), ":nick COMMAND"),
            (Message(None, ("nick", "user"), "COMMAND"), ":nick!user COMMAND"),
            (
                Message(None, ("nick", "user", "host"), "COMMAND"),
                ":nick!user@host COMMAND",
            ),
            (
                Message({"a": "b"}, "nick", "COMMAND", "a", "b"),
                "@a=b :nick COMMAND a b",
            ),
        ],
    )
    def test_str(self, obj: Message, text: str) -> None:
        """Test string conversion"""
        assert str(obj) == text

    @pytest.mark.parametrize(
        ("tags", "prefix", "command", "params"),
        [(None, None, "", []), ("some tag", None, "COMMAND", ["param", ""])],
    )
    def test_eq(
        self,
        tags: Optional[str],
        prefix: Optional[str],
        command: str,
        params: List[str],
    ) -> None:
        """Test equals"""
        assert Message(tags, prefix, command, params) == Message(
            tags, prefix, command, params
        )

    @pytest.mark.parametrize(
        ("tags", "prefix", "command", "params"),
        [(None, None, "", []), ("some tag", None, "COMMAND", ["param", ""])],
    )
    def test_ne(
        self,
        tags: Optional[str],
        prefix: Optional[str],
        command: str,
        params: List[str],
    ) -> None:
        """Test not-equals"""
        b = Message(tags, prefix, command, params) != Message(
            tags, prefix, command, params
        )
        assert not b

    @pytest.mark.parametrize(
        ("obj", "other"),
        [
            (Message(None, None, ""), 0),
            (Message(None, None, ""), None),
            (Message(None, None, ""), ()),
        ],
    )
    def test_no_cmp(self, obj: Message, other: object) -> None:
        """Test Message.__ne__"""
        assert obj != other
        assert other != obj

        assert not obj == other
        assert not other == obj

    @pytest.mark.parametrize("obj", [Message(None, None, "COMMAND")])
    def test_bool(self, obj: Message) -> None:
        """Test the cases where bool(Message) should return True"""
        assert obj

    @pytest.mark.parametrize(
        "obj",
        [
            Message(None, None, None),
            Message(None, None, ""),
            Message(None, "", ""),
            Message("", "", ""),
            Message([], [], "", []),
            Message({}, [], "", []),
            Message(TagList(), Prefix(), "", ParamList()),
        ],
    )
    def test_bool_false(self, obj: Message) -> None:
        """Test all the cases where bool(Message) should return False"""
        assert not obj


def test_trail() -> None:
    """
    Ensure this parser does not have the same
    issue as https://github.com/hexchat/hexchat/issues/2271
    """
    text = "COMMAND thing thing :thing"
    parsed = Message.parse(text)

    assert parsed.command == "COMMAND"

    for i in range(3):
        assert parsed.parameters[i] == "thing"


@pytest.mark.parametrize(
    ("parse_type", "text"),
    [
        (MessageTag, "hi"),
        (MessageTag, "hello=world"),
        (MessageTag, "he\\:llo=\\swor\\r\\\\ld"),
        (TagList, "hi"),
        (TagList, "hello; there"),
        (TagList, "hello; there;"),
        (TagList, "hello=world; there"),
        (Prefix, "nick"),
        (Prefix, "nick!user"),
        (Prefix, "nick@host"),
        (Prefix, "nick!user@host"),
        (ParamList, "test test"),
        (ParamList, "test :test"),
        (ParamList, "test :test test"),
        (Message, "COMMAND"),
        (Message, "command"),
    ],
)
def test_comparisons(parse_type: Type[Parseable], text: str) -> None:
    """Test comparing parsed objects to strings"""
    assert text == parse_type.parse(text)
    assert not text != parse_type.parse(text)


@pytest.mark.parametrize("data", parser_tests.data.msg_split["tests"])
def test_msg_split(data: MsgSplitCase) -> None:
    """Test splitting a message against the irc-parser-tests data"""
    msg = Message.parse(data["input"])
    atoms = data["atoms"].copy()

    # We store tags a bit differently than the test data expects, convert the format
    if msg.tags is not None:
        tags_dict = {name: tag.value for name, tag in msg.tags.items()}
    else:
        tags_dict = None

    assert tags_dict == atoms.get("tags", None)

    prefix = None if msg.prefix is None else str(msg.prefix)
    assert prefix == atoms.get("source", None)

    # Commands are case-insensitive
    assert String(msg.command, ASCII) == atoms.get("verb", None)

    assert list(msg.parameters) == atoms.get("params", [])


@pytest.mark.parametrize("data", parser_tests.data.userhost_split["tests"])
def test_userhost_split(data: UserHostSplitCase) -> None:
    """Ensure that user/host parsing passes against the irc-parser-tests data"""
    source = Prefix.parse(data["source"])
    atoms = data["atoms"].copy()

    assert source.nick == atoms.get("nick", "")
    assert source.user == atoms.get("user", "")
    assert source.host == atoms.get("host", "")


def test_message_tag_repr() -> None:
    """Test repr(MessageTag)"""
    m = MessageTag("foo", "bar")
    assert repr(m) == "MessageTag(name='foo', value='bar')"


@pytest.mark.parametrize("data", parser_tests.data.msg_join["tests"])
def test_msg_join(data: MsgJoinCase) -> None:
    """
    Ensure that message building passes all tests from
    the irc-parser-tests library.
    """
    atoms = data["atoms"]
    msg = Message(
        atoms.get("tags", None),
        atoms.get("source", None),
        atoms.get("verb", None),
        atoms.get("params", []),
    )

    matches = data["matches"]
    assert str(msg) in matches


@pytest.mark.parametrize(
    ("text", "has_trail"),
    [
        ("PRIVMSG #channel :message", True),
        ("PRIVMSG #channel :message text", True),
    ],
)
def test_has_trail(text: str, has_trail: bool) -> None:
    """Ensure that a message with trailing arguments is recorded as having a tril"""
    msg = Message.parse(text)
    assert msg.parameters.has_trail == has_trail



"""Test commands util"""

import pytest

from irclib.util import commands
from irclib.util.commands import CommandArgument


def test_command_lookup() -> None:
    """Test looking up a command"""
    pm = commands.client_commands["privmsg"]
    assert pm is commands.client_commands["PRIVMSG"]
    assert pm is commands.client_commands.privmsg
    assert pm is commands.client_commands.PRIVMSG

    with pytest.raises(KeyError):
        _ = commands.client_commands["foo"]


@pytest.mark.parametrize(
    ("text", "name", "required"),
    [("<foo>", "foo", True), ("[foo]", "foo", False)],
)
def test_arg_parse(text: str, name: str, required: bool) -> None:
    """Test parsing an argument string"""
    arg = CommandArgument.parse(text)
    assert arg.name == name
    assert arg.required is required


def test_parse_error() -> None:
    """Test errors from CommandArgument.parse"""
    with pytest.raises(ValueError, match=r"Unable to parse argument: .*"):
        CommandArgument.parse("{a|b}")



"""Test frozendict util"""

from irclib.util.frozendict import FrozenDict


def test_dict() -> None:
    """Test basic dict functions"""
    fd = FrozenDict(a=1, b=2)
    assert fd["a"] == 1
    assert fd["b"] == 2

    assert len(fd) == 2
    assert set(fd.keys()) == {"a", "b"}


def test_init_literal() -> None:
    """Test initializing with a dict literal"""
    fd = FrozenDict({"a": 1, "b": 2})
    assert fd["a"] == 1
    assert fd["b"] == 2

    assert len(fd) == 2
    assert set(fd.keys()) == {"a", "b"}


def test_copy() -> None:
    """Test dict copy"""
    fd = FrozenDict([("a", 1), ("b", 2)])
    fd1 = fd.copy()
    assert len(fd1) == 2
    assert fd1["a"] == 1
    assert fd1["b"] == 2

    fd2 = fd.copy(c=3)
    assert fd2["a"] == 1
    assert fd2["b"] == 2
    assert fd2["c"] == 3
    assert len(fd2) == 3

    fd3 = fd.copy(a=3)
    assert fd3["a"] == 3
    assert fd3["b"] == 2


def test_hash() -> None:
    """Test dict hashes"""
    fd = FrozenDict(a=1, b=2)
    fd1 = FrozenDict({"a": 1, "b": 2})
    assert fd == fd1
    assert hash(fd) == hash(fd1)
    h = hash(fd)
    h1 = hash(fd)
    assert h == h1



"""Test IRC string class"""

import pytest

from irclib.util.string import ASCII, RFC1459, String


def test_comparisons() -> None:
    """Test comparison logic"""
    # pylint: disable=misplaced-comparison-constant, unneeded-not
    str1 = String("A", ASCII)
    str1_ = str1
    str2 = String("a", ASCII)

    assert str1.casemap is ASCII
    assert str2.casemap is ASCII

    assert str1 == "A"
    assert str2 == "a"

    assert str(str1.lower()) == "a"
    assert str(str2.lower()) == "a"

    assert str(str1.upper()) == "A"
    assert str(str2.upper()) == "A"

    assert str1 == str1_
    assert str1 == str2

    assert str1 == "a"
    assert "a" == str1

    assert str2 == "a"
    assert "a" == str2

    assert str1 == "A"
    assert "A" == str1

    assert str2 == "A"
    assert "A" == str2

    assert not str1 > str2
    assert not str1 > "a"

    assert "B" > str2
    assert not str2 > "B"
    assert str2 < "B"
    assert not "B" < str2
    assert not str2 >= "B"
    assert str2 <= "B"
    assert "B" >= str2
    assert not "B" <= str2

    assert str2 != "B"

    assert str1 != 5


def test_getitem() -> None:
    """Test `getitem`"""
    s = String("abc")
    assert s[0] == "a"
    assert isinstance(s[0], String)
    assert s[0].casemap is s.casemap


def test_capitalize() -> None:
    """Test `capitalize`"""
    s = String("abC")
    assert s.capitalize() == "Abc"


def test_swapcase() -> None:
    """Test `swapcase`"""
    s = String("abC")
    assert s.swapcase() == "ABc"


def test_startswith() -> None:
    """Test `startswith`"""
    s = String("aBc")
    assert s.startswith("a")
    assert s.startswith("ab")
    assert s.startswith(String("aB"))
    assert s.startswith(("a", "b"))


def test_endswith() -> None:
    """Test `endswith`"""
    s = String("aBc")
    assert s.endswith("c")
    assert s.endswith("bC")
    assert s.endswith(String("bc"))
    assert s.endswith(("b", "c"))


def test_find() -> None:
    """Test `find`"""
    s = String("AbCB")
    assert s.find("B") == 1
    assert s.find(String("B")) == 1
    assert s.find("d") == -1


def test_rfind() -> None:
    """Test `rfind`"""
    s = String("AbCB")
    assert s.rfind("B") == 3
    assert s.rfind(String("B")) == 3
    assert s.rfind("d") == -1


def test_index() -> None:
    """Test `index`"""
    s = String("AbCB")
    assert s.index("B") == 1
    assert s.index(String("B")) == 1
    with pytest.raises(ValueError, match="substring not found"):
        s.index("d")


def test_rindex() -> None:
    """Test `rindex`"""
    s = String("AbCB")
    assert s.rindex("B") == 3
    assert s.rindex(String("B")) == 3
    with pytest.raises(ValueError, match="substring not found"):
        s.rindex("d")


def test_count() -> None:
    """Test `count`"""
    assert String("aabbcc").count("a") == 2
    assert String("aabbcc").count(String("a")) == 2


def test_hash() -> None:
    """Test hashing a string"""
    assert hash(String("ABC")) == hash(String("abc"))


def test_contains() -> None:
    """Test `in` operator"""
    assert "a" in String("abc")
    assert String("a") in String("abc")
    assert 445 not in String("abc")


def test_instance() -> None:
    """Ensure that basic strings aren't being counted as String"""
    assert not isinstance("a", String)


def test_rfc1459() -> None:
    """Test rfc1459 casemap"""
    str1 = String("ABC|{", RFC1459)
    str2 = String("abc\\[", RFC1459)

    assert str1 == str2
    assert str(str1.lower()) == str(str2.lower())
    assert str(str1.upper()) == str(str2.upper())


def test_partition() -> None:
    """Test partition"""
    s = String("ABc")
    assert s.partition("b") == ("A", "B", "c")
    assert s.partition("d") == ("ABc", "", "")


def test_rpartition() -> None:
    """Test rpartition"""
    s = String("AbCbAcBa")
    assert s.rpartition("b") == ("AbCbAc", "B", "a")
    assert s.rpartition("d") == ("", "", "AbCbAcBa")


def test_strip() -> None:
    """Test stripping characters"""
    s = String("AbCbAcBa")
    assert s.strip("d") == "AbCbAcBa"
    assert s.strip("a") == "bCbAcB"
    s1 = String(" ABc  ")
    assert s1.strip() == "ABc"



"""Test IRC string comparisons"""

from typing import List, TypedDict

import parser_tests.data
import pytest

from irclib.util.compare import match_mask


class MaskCase(TypedDict):
    mask: str
    matches: List[str]
    fails: List[str]


@pytest.mark.parametrize("data", parser_tests.data.mask_match["tests"])
def test_mask_match(data: MaskCase) -> None:
    """Test the mask matching logic"""
    pattern = data["mask"]

    for hostmask in data["matches"]:
        assert match_mask(hostmask, pattern)

    for hostmask in data["fails"]:
        assert not match_mask(hostmask, pattern)



"""Test numerics util"""

from irclib.util import numerics


def test_lookup() -> None:
    """Test looking up a numeric"""
    n = numerics.numerics["001"]
    assert n is numerics.numerics.from_int(1)
    assert n is numerics.numerics.from_name("RPL_WELCOME")


def test_iter() -> None:
    """Test iterating the numerics list"""
    assert next(iter(numerics.numerics)) == "001"


def test_len() -> None:
    """Test checking the length of the numerics list"""
    assert len(numerics.numerics) > 0
